# Copyrights 2007-2014 by [Mark Overmeer].
#  For other contributors see ChangeLog.
# See the manual pages for details on the licensing terms.
# Pod stripped from pm file by OODoc 2.01.

   die "XML::Compile::Operation got renamed into XML::Compile::SOAP::Operation n release 2.18";
   1;
